# This file is part of pipe_base.
#
# Developed for the LSST Data Management System.
# This product includes software developed by the LSST Project
# (http://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Collection of common methods for use in unit tests."""

from __future__ import annotations

from lsst.daf.butler import DatasetRef

from ..graph import QuantumGraph


def check_output_run(graph: QuantumGraph, run: str) -> list[DatasetRef]:
    """Check that all output and intermediate datasets belong to a
    specified run.

    Parameters
    ----------
    graph : `QuantumGraph`
        Quantum graph.
    run : `str`
        Output run name.

    Returns
    -------
    refs : `list` [ `~lsst.daf.butler.DatasetRef` ]
        List of output/intermediate dataset references that do NOT belong to
        the specified run.
    """
    # Collect all inputs/outputs, so that we can build intermediate refs.
    output_refs = []
    input_refs = []
    for node in graph:
        for refs in node.quantum.outputs.values():
            output_refs += refs
        for refs in node.quantum.inputs.values():
            input_refs += refs
    for task_def in graph.iterTaskGraph():
        init_refs = graph.initOutputRefs(task_def)
        if init_refs:
            output_refs += init_refs
        init_refs = graph.initInputRefs(task_def)
        if init_refs:
            input_refs += init_refs
    output_refs += graph.globalInitOutputRefs()
    refs = [ref for ref in output_refs if ref.run != run]

    output_ids = {ref.id for ref in output_refs}
    intermediates = [ref for ref in input_refs if ref.id in output_ids]
    refs += [ref for ref in intermediates if ref.run != run]

    return refs



#
# LSST Data Management System
# Copyright 2008-2016 AURA/LSST.
#
# This product includes software developed by the
# LSST Project (http://www.lsst.org/).
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the LSST License Statement and
# the GNU General Public License along with this program.  If not,
# see <http://www.lsstcorp.org/LegalNotices/>.
#

from __future__ import annotations

__all__ = ["Task", "TaskError"]

import contextlib
import logging
import weakref
from collections.abc import Callable, Iterator, Sequence
from typing import TYPE_CHECKING, Any, ClassVar

import lsst.utils
import lsst.utils.logging
from lsst.pex.config import ConfigurableField
from lsst.utils.timer import logInfo

if TYPE_CHECKING:
    from lsst.pex.config import Config

try:
    import lsstDebug  # type: ignore
except ImportError:
    lsstDebug = None

from ._task_metadata import TaskMetadata

# This defines the Python type to use for task metadata. It is a private
# class variable that can be accessed by other closely-related middleware
# code and test code.
_TASK_METADATA_TYPE = TaskMetadata
_TASK_FULL_METADATA_TYPE = TaskMetadata


class TaskError(Exception):
    """Use to report errors for which a traceback is not useful.

    Notes
    -----
    Examples of such errors:

    - processCcd is asked to run detection, but not calibration, and no calexp
      is found.
    - coadd finds no valid images in the specified patch.
    """

    pass


class Task:
    r"""Base class for data processing tasks.

    See :ref:`task-framework-overview` to learn what tasks are, and
    :ref:`creating-a-task` for more information about writing tasks.

    Parameters
    ----------
    config : `Task.ConfigClass` instance, optional
        Configuration for this task (an instance of Task.ConfigClass, which
        is a task-specific subclass of `lsst.pex.config.Config`, or `None`.
        If `None`:

        - If parentTask specified then defaults to parentTask.config.\<name>
        - If parentTask is None then defaults to self.ConfigClass()

    name : `str`, optional
        Brief name of task, or `None`; if `None` then defaults to
        `Task._DefaultName`
    parentTask : `Task`-type, optional
        The parent task of this subtask, if any.

        - If `None` (a top-level task) then you must specify config and name
          is ignored.
        - If not `None` (a subtask) then you must specify name.
    log : `logging.Logger` or subclass, optional
        Log whose name is used as a log name prefix, or `None` for no prefix.
        Ignored if is parentTask specified, in which case
        ``parentTask.log``\ 's name is used as a prefix. The task's log name is
        ``prefix + "." + name`` if a prefix exists, else ``name``. The task's
        log is then a child logger of ``parentTask.log`` (if ``parentTask``
        specified), or a child logger of the log from the argument
        (if ``log`` is not `None`).

    Raises
    ------
    RuntimeError
        Raised under these circumstances:

        - If ``parentTask`` is `None` and ``config`` is `None`.
        - If ``parentTask`` is not `None` and ``name`` is `None`.
        - If ``name`` is `None` and ``_DefaultName`` does not exist.

    Notes
    -----
    The constructor must use keyword parameters for everything other than
    the ``config`` parameter which can be positional or use keyword form.

    Useful attributes include:

    - ``log``: an `logging.Logger` or subclass.
    - ``config``: task-specific configuration; an instance of ``ConfigClass``
      (see below).
    - ``metadata``: a `TaskMetadata` for
      collecting task-specific metadata, e.g. data quality and performance
      metrics. This is data that is only meant to be persisted, never to be
      used by the task.

    Use a `lsst.pipe.base.PipelineTask` subclass to perform I/O with a
    Butler.

    Subclasses must also have an attribute ``ConfigClass`` that is a subclass
    of `lsst.pex.config.Config` which configures the task. Subclasses should
    also have an attribute ``_DefaultName``: the default name if there is no
    parent task. ``_DefaultName`` is required for subclasses of
    `~lsst.pipe.base.PipeLineTask` and recommended for subclasses of Task
    because it simplifies construction (e.g. for unit tests).
    """

    ConfigClass: ClassVar[type[Config]]
    _DefaultName: ClassVar[str]

    _add_module_logger_prefix: bool = True
    """Control whether the module prefix should be prepended to default
    logger names."""

    def __init__(
        self,
        config: Config | None = None,
        *,
        name: str | None = None,
        parentTask: Task | None = None,
        log: logging.Logger | lsst.utils.logging.LsstLogAdapter | None = None,
    ):
        self.metadata = _TASK_METADATA_TYPE()
        self.__parentTask: weakref.ReferenceType | None
        self.__parentTask = parentTask if parentTask is None else weakref.ref(parentTask)

        if parentTask is not None:
            if name is None:
                raise RuntimeError("name is required for a subtask")
            self._name = name
            self._fullName = parentTask._computeFullName(name)
            if config is None:
                config = getattr(parentTask.config, name)
            self._taskDict: dict[str, weakref.ReferenceType[Task]] = parentTask._taskDict
            loggerName = parentTask.log.getChild(name).name
        else:
            if name is None:
                name = getattr(self, "_DefaultName", None)
                if name is None:
                    raise RuntimeError("name is required for a task unless it has attribute _DefaultName")
                name = self._DefaultName
            self._name = name
            self._fullName = self._name
            if config is None:
                config = self.ConfigClass()
            self._taskDict = dict()
            loggerName = self._fullName
            if log is not None and log.name:
                loggerName = log.getChild(loggerName).name
            elif self._add_module_logger_prefix:
                # Prefix the logger name with the root module name.
                # We want all Task loggers to have this prefix to make
                # it easier to control them. This can be disabled by
                # a Task setting the class property _add_module_logger_prefix
                # to False -- in which case the logger name will not be
                # modified.
                module_name = self.__module__
                module_root = module_name.split(".")[0] + "."
                if not loggerName.startswith(module_root):
                    loggerName = module_root + loggerName

        # Get a logger (that might be a subclass of logging.Logger).
        self.log: lsst.utils.logging.LsstLogAdapter = lsst.utils.logging.getLogger(loggerName)
        self.config: Config = config
        self.config.validate()
        if lsstDebug:
            self._display = lsstDebug.Info(self.__module__).display
        else:
            self._display = None
        self._taskDict[self._fullName] = weakref.ref(self)

    @property
    def _parentTask(self) -> Task | None:
        return self.__parentTask if self.__parentTask is None else self.__parentTask()

    def emptyMetadata(self) -> None:
        """Empty (clear) the metadata for this Task and all sub-Tasks."""
        for wref in self._taskDict.values():
            subtask = wref()
            assert subtask is not None, "Unexpected garbage collection of subtask."
            subtask.metadata = _TASK_METADATA_TYPE()

    def getFullMetadata(self) -> TaskMetadata:
        """Get metadata for all tasks.

        Returns
        -------
        metadata : `TaskMetadata`
            The keys are the full task name.
            Values are metadata for the top-level task and all subtasks,
            sub-subtasks, etc.

        Notes
        -----
        The returned metadata includes timing information (if
        ``@timer.timeMethod`` is used) and any metadata set by the task. The
        name of each item consists of the full task name with ``.`` replaced
        by ``:``, followed by ``.`` and the name of the item, e.g.::

            topLevelTaskName:subtaskName:subsubtaskName.itemName

        using ``:`` in the full task name disambiguates the rare situation
        that a task has a subtask and a metadata item with the same name.
        """
        fullMetadata = _TASK_FULL_METADATA_TYPE()
        for fullName, wref in self.getTaskDict().items():
            subtask = wref()
            assert subtask is not None, "Unexpected garbage collection of subtask."
            fullMetadata[fullName.replace(".", ":")] = subtask.metadata
        return fullMetadata

    def getFullName(self) -> str:
        """Get the task name as a hierarchical name including parent task
        names.

        Returns
        -------
        fullName : `str`
            The full name consists of the name of the parent task and each
            subtask separated by periods. For example:

            - The full name of top-level task "top" is simply "top".
            - The full name of subtask "sub" of top-level task "top" is
              "top.sub".
            - The full name of subtask "sub2" of subtask "sub" of top-level
              task "top" is "top.sub.sub2".
        """
        return self._fullName

    def getName(self) -> str:
        """Get the name of the task.

        Returns
        -------
        taskName : `str`
            Name of the task.

        See Also
        --------
        getFullName
        """
        return self._name

    def getTaskDict(self) -> dict[str, weakref.ReferenceType[Task]]:
        """Get a dictionary of all tasks as a shallow copy.

        Returns
        -------
        taskDict : `dict`
            Dictionary containing full task name: task object for the top-level
            task and all subtasks, sub-subtasks, etc.
        """
        return self._taskDict.copy()

    def makeSubtask(self, name: str, **keyArgs: Any) -> None:
        """Create a subtask as a new instance as the ``name`` attribute of this
        task.

        Parameters
        ----------
        name : `str`
            Brief name of the subtask.
        **keyArgs
            Extra keyword arguments used to construct the task. The following
            arguments are automatically provided and cannot be overridden:

            - ``config``.
            - ``parentTask``.

        Notes
        -----
        The subtask must be defined by ``Task.config.name``, an instance of
        `~lsst.pex.config.ConfigurableField` or
        `~lsst.pex.config.RegistryField`.
        """
        taskField = getattr(self.config, name, None)
        if taskField is None:
            raise KeyError(f"{self.getFullName()}'s config does not have field {name!r}")
        subtask = taskField.apply(name=name, parentTask=self, **keyArgs)
        setattr(self, name, subtask)

    @contextlib.contextmanager
    def timer(self, name: str, logLevel: int = logging.DEBUG) -> Iterator[None]:
        """Context manager to log performance data for an arbitrary block of
        code.

        Parameters
        ----------
        name : `str`
            Name of code being timed; data will be logged using item name:
            ``Start`` and ``End``.
        logLevel
            A `logging` level constant.

        Examples
        --------
        Creating a timer context:

        .. code-block:: python

            with self.timer("someCodeToTime"):
                pass  # code to time

        See Also
        --------
        lsst.utils.timer.logInfo
        """
        logInfo(obj=self, prefix=name + "Start", logLevel=logLevel)
        try:
            yield
        finally:
            logInfo(obj=self, prefix=name + "End", logLevel=logLevel)

    @classmethod
    def makeField(cls, doc: str) -> ConfigurableField:
        """Make a `lsst.pex.config.ConfigurableField` for this task.

        Parameters
        ----------
        doc : `str`
            Help text for the field.

        Returns
        -------
        configurableField : `lsst.pex.config.ConfigurableField`
            A `~lsst.pex.config.ConfigurableField` for this task.

        Examples
        --------
        Provides a convenient way to specify this task is a subtask of another
        task.

        Here is an example of use:

        .. code-block:: python

            class OtherTaskConfig(lsst.pex.config.Config):
                aSubtask = ATaskClass.makeField("brief description of task")
        """
        return ConfigurableField(doc=doc, target=cls)

    def _computeFullName(self, name: str) -> str:
        """Compute the full name of a subtask or metadata item, given its brief
        name.

        Parameters
        ----------
        name : `str`
            Brief name of subtask or metadata item.

        Returns
        -------
        fullName : `str`
            The full name: the ``name`` argument prefixed by the full task name
            and a period.

        Notes
        -----
        For example: if the full name of this task is "top.sub.sub2"
        then ``_computeFullName("subname")`` returns
        ``"top.sub.sub2.subname"``.
        """
        return f"{self._fullName}.{name}"

    @staticmethod
    def _unpickle_via_factory(
        factory: Callable[..., Task], args: Sequence[Any], kwargs: dict[str, Any]
    ) -> Task:
        """Unpickle something by calling a factory

        Allows subclasses to unpickle using `__reduce__` with keyword
        arguments as well as positional arguments.
        """
        return factory(*args, **kwargs)

    def _reduce_kwargs(self) -> dict[str, Any]:
        """Return a dict of the keyword arguments that should be used
        by `__reduce__`.

        Subclasses with additional arguments should always call the parent
        class method to ensure that the standard parameters are included.

        Returns
        -------
        kwargs : `dict`
            Keyword arguments to be used when pickling.
        """
        return dict(
            config=self.config,
            name=self._name,
            parentTask=self._parentTask,
        )

    def __reduce__(
        self,
    ) -> tuple[
        Callable[[Callable[..., Task], Sequence[Any], dict[str, Any]], Task],
        tuple[type[Task], Sequence[Any], dict[str, Any]],
    ]:
        """Pickler."""
        return self._unpickle_via_factory, (self.__class__, [], self._reduce_kwargs())



# This file is part of pipe_base.
#
# Developed for the LSST Data Management System.
# This product includes software developed by the LSST Project
# (https://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""Tests of the Instrument class.
"""

import datetime
import math
import unittest

from lsst.daf.butler import DataCoordinate, DimensionPacker, DimensionUniverse, Registry, RegistryConfig
from lsst.daf.butler.formatters.json import JsonFormatter
from lsst.pex.config import Config
from lsst.pipe.base import Instrument
from lsst.utils.introspection import get_full_type_name


class DummyInstrument(Instrument):
    @classmethod
    def getName(cls):
        return "DummyInstrument"

    def register(self, registry, update=False):
        detector_max = 2
        visit_max = 10
        exposure_max = 8
        record = {
            "instrument": self.getName(),
            "class_name": get_full_type_name(DummyInstrument),
            "detector_max": detector_max,
            "visit_max": visit_max,
            "exposure_max": exposure_max,
        }
        with registry.transaction():
            registry.syncDimensionData("instrument", record, update=update)

    def getRawFormatter(self, dataId):
        return JsonFormatter


class BadInstrument(DummyInstrument):
    """Instrument with wrong class name."""

    raw_definition = ("raw2", ("instrument", "detector", "exposure"), "StructuredDataDict")

    @classmethod
    def getName(cls):
        return "BadInstrument"

    def register(self, registry, update=False):
        # Register a bad class name
        record = {
            "instrument": self.getName(),
            "class_name": "builtins.str",
            "detector_max": 1,
        }
        registry.syncDimensionData("instrument", record, update=update)


class UnimportableInstrument(DummyInstrument):
    """Instrument with class name that does not exist."""

    @classmethod
    def getName(cls):
        return "NoImportInstr"

    def register(self, registry, update=False):
        # Register a bad class name
        record = {
            "instrument": self.getName(),
            "class_name": "not.importable",
            "detector_max": 1,
        }
        registry.syncDimensionData("instrument", record, update=update)


class DimensionPackerTestConfig(Config):
    packer = Instrument.make_dimension_packer_config_field()


class InstrumentTestCase(unittest.TestCase):
    """Test for Instrument."""

    def setUp(self):
        self.instrument = DummyInstrument()
        self.name = "DummyInstrument"

    def test_basics(self):
        self.assertEqual(self.instrument.getName(), self.name)
        self.assertEqual(self.instrument.getRawFormatter({}), JsonFormatter)
        self.assertIsNone(DummyInstrument.raw_definition)
        raw = BadInstrument.raw_definition
        self.assertEqual(raw[2], "StructuredDataDict")

    def test_register(self):
        """Test that register() sets appropriate Dimensions."""
        registryConfig = RegistryConfig()
        registryConfig["db"] = "sqlite://"
        registry = Registry.createFromConfig(registryConfig)
        # Check that the registry starts out empty.
        self.instrument.importAll(registry)
        self.assertFalse(list(registry.queryDimensionRecords("instrument")))

        # Register and check again.
        self.instrument.register(registry)
        instruments = list(registry.queryDimensionRecords("instrument"))
        self.assertEqual(len(instruments), 1)
        self.assertEqual(instruments[0].name, self.name)
        self.assertEqual(instruments[0].detector_max, 2)
        self.assertIn("DummyInstrument", instruments[0].class_name)

        self.instrument.importAll(registry)
        from_registry = DummyInstrument.fromName("DummyInstrument", registry)
        self.assertIsInstance(from_registry, Instrument)
        with self.assertRaises(LookupError):
            Instrument.fromName("NotThrere", registry)

        # Register a bad instrument.
        BadInstrument().register(registry)
        with self.assertRaises(TypeError):
            Instrument.fromName("BadInstrument", registry)

        UnimportableInstrument().register(registry)
        with self.assertRaises(ImportError):
            Instrument.fromName("NoImportInstr", registry)

        # This should work even with the bad class name.
        self.instrument.importAll(registry)

    def test_defaults(self):
        self.assertEqual(self.instrument.makeDefaultRawIngestRunName(), "DummyInstrument/raw/all")
        self.assertEqual(
            self.instrument.makeUnboundedCalibrationRunName("a", "b"), "DummyInstrument/calib/a/b/unbounded"
        )
        self.assertEqual(
            self.instrument.makeCuratedCalibrationRunName("2018-05-04", "a"),
            "DummyInstrument/calib/a/curated/20180504T000000Z",
        )
        self.assertEqual(self.instrument.makeCalibrationCollectionName("c"), "DummyInstrument/calib/c")
        self.assertEqual(self.instrument.makeRefCatCollectionName(), "refcats")
        self.assertEqual(self.instrument.makeRefCatCollectionName("a"), "refcats/a")
        self.assertEqual(self.instrument.makeUmbrellaCollectionName(), "DummyInstrument/defaults")

        instrument = DummyInstrument(collection_prefix="Different")
        self.assertEqual(instrument.makeCollectionName("a"), "Different/a")
        self.assertEqual(self.instrument.makeCollectionName("a"), "DummyInstrument/a")

    def test_collection_timestamps(self):
        self.assertEqual(
            Instrument.formatCollectionTimestamp("2018-05-03"),
            "20180503T000000Z",
        )
        self.assertEqual(
            Instrument.formatCollectionTimestamp("2018-05-03T14:32:16"),
            "20180503T143216Z",
        )
        self.assertEqual(
            Instrument.formatCollectionTimestamp("20180503T143216Z"),
            "20180503T143216Z",
        )
        self.assertEqual(
            Instrument.formatCollectionTimestamp(datetime.datetime(2018, 5, 3, 14, 32, 16)),
            "20180503T143216Z",
        )
        formattedNow = Instrument.makeCollectionTimestamp()
        self.assertIsInstance(formattedNow, str)
        datetimeThen1 = datetime.datetime.strptime(formattedNow, "%Y%m%dT%H%M%S%z")
        self.assertEqual(datetimeThen1.tzinfo, datetime.timezone.utc)

        with self.assertRaises(TypeError):
            Instrument.formatCollectionTimestamp(0)

    def test_dimension_packer_config_defaults(self):
        """Test the ObservationDimensionPacker class and the Instrument-based
        systems for constructing it, in the case where the Instrument-defined
        default is used.
        """
        registry_config = RegistryConfig()
        registry_config["db"] = "sqlite://"
        registry = Registry.createFromConfig(registry_config)
        self.instrument.register(registry)
        config = DimensionPackerTestConfig()
        instrument_data_id = registry.expandDataId(instrument=self.name)
        record = instrument_data_id.records["instrument"]
        self.check_dimension_packers(
            registry.dimensions,
            # Test just one packer-construction signature here as that saves us
            # from having to insert dimension records for visits, exposures,
            # and detectors for this test.
            # Note that we don't need to pass any more than the instrument in
            # the data ID yet, because we're just constructing packers, not
            # calling their pack method.
            visit_packers=[config.packer.apply(instrument_data_id, is_exposure=False)],
            exposure_packers=[config.packer.apply(instrument_data_id, is_exposure=True)],
            n_detectors=record.detector_max,
            n_visits=record.visit_max,
            n_exposures=record.exposure_max,
        )

    def test_dimension_packer_config_override(self):
        """Test the ObservationDimensionPacker class and the Instrument-based
        systems for constructing it, in the case where configuration overrides
        the Instrument's default.
        """
        registry_config = RegistryConfig()
        registry_config["db"] = "sqlite://"
        registry = Registry.createFromConfig(registry_config)
        # Intentionally do not register instrument or insert any other
        # dimension records to ensure we don't need them in this mode.
        config = DimensionPackerTestConfig()
        config.packer["observation"].n_observations = 16
        config.packer["observation"].n_detectors = 4
        config.packer.name = "observation"
        instrument_data_id = DataCoordinate.standardize(instrument=self.name, universe=registry.dimensions)
        full_data_id = DataCoordinate.standardize(instrument_data_id, detector=1, visit=3, exposure=7)
        visit_data_id = full_data_id.subset(full_data_id.universe.extract(["visit", "detector"]))
        exposure_data_id = full_data_id.subset(full_data_id.universe.extract(["exposure", "detector"]))
        self.check_dimension_packers(
            registry.dimensions,
            # Note that we don't need to pass any more than the instrument in
            # the data ID yet, because we're just constructing packers, not
            # calling their pack method.
            visit_packers=[
                config.packer.apply(visit_data_id),
                config.packer.apply(full_data_id),
                config.packer.apply(instrument_data_id, is_exposure=False),
            ],
            exposure_packers=[
                config.packer.apply(instrument_data_id, is_exposure=True),
                config.packer.apply(exposure_data_id),
            ],
            n_detectors=config.packer["observation"].n_detectors,
            n_visits=config.packer["observation"].n_observations,
            n_exposures=config.packer["observation"].n_observations,
        )

    def check_dimension_packers(
        self,
        universe: DimensionUniverse,
        visit_packers: list[DimensionPacker],
        exposure_packers: list[DimensionPacker],
        n_detectors: int,
        n_visits: int,
        n_exposures: int,
    ) -> None:
        """Test the behavior of one or more dimension packers constructed by
        an instrument.

        Parameters
        ----------
        universe : `lsst.daf.butler.DimensionUniverse`
            Data model for butler data IDs.
        visit_packers : `list` [ `DimensionPacker` ]
            Packers with ``{visit, detector}`` dimensions to test.
        exposure_packers : `list` [ `DimensionPacker` ]
            Packers with ``{exposure, detector}`` dimensions to test.
        n_detectors : `int`
            Number of detectors all packers have been configured to reserve
            space for.
        n_visits : `int`
            Number of visits all packers in ``visit_packers`` have been
            configured to reserve space for.
        n_exposures : `int`
            Number of exposures all packers in ``exposure_packers`` have been
            configured to reserve space for.
        """
        full_data_id = DataCoordinate.standardize(
            instrument=self.name, detector=1, visit=3, exposure=7, universe=universe
        )
        visit_data_id = full_data_id.subset(full_data_id.universe.extract(["visit", "detector"]))
        exposure_data_id = full_data_id.subset(full_data_id.universe.extract(["exposure", "detector"]))
        for n, packer in enumerate(visit_packers):
            with self.subTest(n=n):
                packed_id, max_bits = packer.pack(full_data_id, returnMaxBits=True)
                self.assertEqual(packed_id, full_data_id["detector"] + full_data_id["visit"] * n_detectors)
                self.assertEqual(max_bits, math.ceil(math.log2(n_detectors * n_visits)))
                self.assertEqual(visit_data_id, packer.unpack(packed_id))
                with self.assertRaisesRegex(ValueError, f"Detector ID {n_detectors} is out of bounds"):
                    packer.pack(instrument=self.name, detector=n_detectors, visit=3)
                with self.assertRaisesRegex(ValueError, f"Visit ID {n_visits} is out of bounds"):
                    packer.pack(instrument=self.name, detector=1, visit=n_visits)
        for n, packer in enumerate(exposure_packers):
            with self.subTest(n=n):
                packed_id, max_bits = packer.pack(full_data_id, returnMaxBits=True)
                self.assertEqual(packed_id, full_data_id["detector"] + full_data_id["exposure"] * n_detectors)
                self.assertEqual(max_bits, math.ceil(math.log2(n_detectors * n_exposures)))
                self.assertEqual(exposure_data_id, packer.unpack(packed_id))
                with self.assertRaisesRegex(ValueError, f"Detector ID {n_detectors} is out of bounds"):
                    packer.pack(instrument=self.name, detector=n_detectors, exposure=3)
                with self.assertRaisesRegex(ValueError, f"Exposure ID {n_exposures} is out of bounds"):
                    packer.pack(instrument=self.name, detector=1, exposure=n_exposures)


if __name__ == "__main__":
    unittest.main()



# This file is part of pipe_base.
#
# Developed for the LSST Data Management System.
# This product includes software developed by the LSST Project
# (http://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
from __future__ import annotations

__all__ = ("MockPipelineTask", "MockPipelineTaskConfig", "mock_task_defs")

import dataclasses
import logging
from collections.abc import Iterable, Mapping
from typing import TYPE_CHECKING, Any, ClassVar

from lsst.daf.butler import DeferredDatasetHandle
from lsst.pex.config import ConfigurableField, Field, ListField
from lsst.utils.doImport import doImportType
from lsst.utils.introspection import get_full_type_name
from lsst.utils.iteration import ensure_iterable

from ...config import PipelineTaskConfig
from ...connections import InputQuantizedConnection, OutputQuantizedConnection, PipelineTaskConnections
from ...pipeline import TaskDef
from ...pipelineTask import PipelineTask
from ._data_id_match import DataIdMatch
from ._storage_class import MockDataset, MockDatasetQuantum, MockStorageClass, get_mock_name

_LOG = logging.getLogger(__name__)

if TYPE_CHECKING:
    from ...butlerQuantumContext import ButlerQuantumContext


def mock_task_defs(
    originals: Iterable[TaskDef],
    unmocked_dataset_types: Iterable[str] = (),
    force_failures: Mapping[str, tuple[str, type[Exception]]] | None = None,
) -> list[TaskDef]:
    """Create mocks for an iterable of TaskDefs.

    Parameters
    ----------
    originals : `~collections.abc.Iterable` [ `TaskDef` ]
        Original tasks and configuration to mock.
    unmocked_dataset_types : `~collections.abc.Iterable` [ `str` ], optional
        Names of overall-input dataset types that should not be replaced with
        mocks.
    force_failures : `~collections.abc.Mapping` [ `str`, `tuple` [ `str`, \
            `type` [ `Exception` ] ] ]
        Mapping from original task label to a 2-tuple indicating that some
        quanta should raise an exception when executed.  The first entry is a
        data ID match using the butler expression language (i.e. a string of
        the sort passed ass the ``where`` argument to butler query methods),
        while the second is the type of exception to raise when the quantum
        data ID matches the expression.

    Returns
    -------
    mocked : `list` [ `TaskDef` ]
        List of `TaskDef` objects using `MockPipelineTask` configurations that
        target the original tasks, in the same order.
    """
    unmocked_dataset_types = tuple(unmocked_dataset_types)
    if force_failures is None:
        force_failures = {}
    results: list[TaskDef] = []
    for original_task_def in originals:
        config = MockPipelineTaskConfig()
        config.original.retarget(original_task_def.taskClass)
        config.original = original_task_def.config
        config.unmocked_dataset_types.extend(unmocked_dataset_types)
        if original_task_def.label in force_failures:
            condition, exception_type = force_failures[original_task_def.label]
            config.fail_condition = condition
            config.fail_exception = get_full_type_name(exception_type)
        mock_task_def = TaskDef(
            config=config, taskClass=MockPipelineTask, label=get_mock_name(original_task_def.label)
        )
        results.append(mock_task_def)
    return results


class MockPipelineDefaultTargetConnections(PipelineTaskConnections, dimensions=()):
    pass


class MockPipelineDefaultTargetConfig(
    PipelineTaskConfig, pipelineConnections=MockPipelineDefaultTargetConnections
):
    pass


class MockPipelineDefaultTargetTask(PipelineTask):
    """A `PipelineTask` class used as the default target for
    ``MockPipelineTaskConfig.original``.

    This is effectively a workaround for `lsst.pex.config.ConfigurableField`
    not supporting ``optional=True``, but that is generally a reasonable
    limitation for production code and it wouldn't make sense just to support
    test utilities.
    """

    ConfigClass = MockPipelineDefaultTargetConfig


class MockPipelineTaskConnections(PipelineTaskConnections, dimensions=()):
    def __init__(self, *, config: MockPipelineTaskConfig):
        original: PipelineTaskConnections = config.original.connections.ConnectionsClass(
            config=config.original.value
        )
        self.dimensions.update(original.dimensions)
        unmocked_dataset_types = frozenset(config.unmocked_dataset_types)
        for name, connection in original.allConnections.items():
            if name in original.initInputs or name in original.initOutputs:
                # We just ignore initInputs and initOutputs, because the task
                # is never given DatasetRefs for those and hence can't create
                # mocks.
                continue
            if connection.name not in unmocked_dataset_types:
                # We register the mock storage class with the global singleton
                # here, but can only put its name in the connection. That means
                # the same global singleton (or one that also has these
                # registrations) has to be available whenever this dataset type
                # is used.
                storage_class = MockStorageClass.get_or_register_mock(connection.storageClass)
                kwargs = {}
                if hasattr(connection, "dimensions"):
                    connection_dimensions = set(connection.dimensions)
                    # Replace the generic "skypix" placeholder with htm7, since
                    # that requires the dataset type to have already been
                    # registered.
                    if "skypix" in connection_dimensions:
                        connection_dimensions.remove("skypix")
                        connection_dimensions.add("htm7")
                    kwargs["dimensions"] = connection_dimensions
                connection = dataclasses.replace(
                    connection,
                    name=get_mock_name(connection.name),
                    storageClass=storage_class.name,
                    **kwargs,
                )
            elif name in original.outputs:
                raise ValueError(f"Unmocked dataset type {connection.name!r} cannot be used as an output.")
            setattr(self, name, connection)


class MockPipelineTaskConfig(PipelineTaskConfig, pipelineConnections=MockPipelineTaskConnections):
    """Configuration class for `MockPipelineTask`."""

    fail_condition = Field[str](
        dtype=str,
        default="",
        doc=(
            "Condition on Data ID to raise an exception. String expression which includes attributes of "
            "quantum data ID using a syntax of daf_butler user expressions (e.g. 'visit = 123')."
        ),
    )

    fail_exception = Field[str](
        dtype=str,
        default="builtins.ValueError",
        doc=(
            "Class name of the exception to raise when fail condition is triggered. Can be "
            "'lsst.pipe.base.NoWorkFound' to specify non-failure exception."
        ),
    )

    original: ConfigurableField = ConfigurableField(
        doc="The original task being mocked by this one.", target=MockPipelineDefaultTargetTask
    )

    unmocked_dataset_types = ListField[str](
        doc=(
            "Names of input dataset types that should be used as-is instead "
            "of being mocked.  May include dataset types not relevant for "
            "this task, which will be ignored."
        ),
        default=(),
        optional=False,
    )

    def data_id_match(self) -> DataIdMatch | None:
        if not self.fail_condition:
            return None
        return DataIdMatch(self.fail_condition)


class MockPipelineTask(PipelineTask):
    """Implementation of `PipelineTask` used for running a mock pipeline.

    Notes
    -----
    This class overrides `runQuantum` to read inputs and write a bit of
    provenance into all of its outputs (always `MockDataset` instances).  It
    can also be configured to raise exceptions on certain data IDs.  It reads
    `MockDataset` inputs and simulates reading inputs of other types by
    creating `MockDataset` inputs from their DatasetRefs.

    At present `MockPipelineTask` simply drops any ``initInput`` and
    ``initOutput`` connections present on the original, since `MockDataset`
    creation for those would have to happen in the code that executes the task,
    not in the task itself.  Because `MockPipelineTask` never instantiates the
    mock task (just its connections class), this is a limitation on what the
    mocks can be used to test, not anything deeper.
    """

    ConfigClass: ClassVar[type[PipelineTaskConfig]] = MockPipelineTaskConfig

    def __init__(
        self,
        *,
        config: MockPipelineTaskConfig,
        **kwargs: Any,
    ):
        super().__init__(config=config, **kwargs)
        self.fail_exception: type | None = None
        self.data_id_match = self.config.data_id_match()
        if self.data_id_match:
            self.fail_exception = doImportType(self.config.fail_exception)

    config: MockPipelineTaskConfig

    def runQuantum(
        self,
        butlerQC: ButlerQuantumContext,
        inputRefs: InputQuantizedConnection,
        outputRefs: OutputQuantizedConnection,
    ) -> None:
        # docstring is inherited from the base class
        quantum = butlerQC.quantum

        _LOG.info("Mocking execution of task '%s' on quantum %s", self.getName(), quantum.dataId)

        assert quantum.dataId is not None, "Quantum DataId cannot be None"

        # Possibly raise an exception.
        if self.data_id_match is not None and self.data_id_match.match(quantum.dataId):
            _LOG.info("Simulating failure of task '%s' on quantum %s", self.getName(), quantum.dataId)
            message = f"Simulated failure: task={self.getName()} dataId={quantum.dataId}"
            assert self.fail_exception is not None, "Exception type must be defined"
            raise self.fail_exception(message)

        # Populate the bit of provenance we store in all outputs.
        _LOG.info("Reading input data for task '%s' on quantum %s", self.getName(), quantum.dataId)
        mock_dataset_quantum = MockDatasetQuantum(
            task_label=self.getName(), data_id=quantum.dataId.to_simple(), inputs={}
        )
        for name, refs in inputRefs:
            inputs_list = []
            for ref in ensure_iterable(refs):
                if isinstance(ref.datasetType.storageClass, MockStorageClass):
                    input_dataset = butlerQC.get(ref)
                    if isinstance(input_dataset, DeferredDatasetHandle):
                        input_dataset = input_dataset.get()
                    if not isinstance(input_dataset, MockDataset):
                        raise TypeError(
                            f"Expected MockDataset instance for {ref}; "
                            f"got {input_dataset!r} of type {type(input_dataset)!r}."
                        )
                    # To avoid very deep provenance we trim inputs to a single
                    # level.
                    input_dataset.quantum = None
                else:
                    input_dataset = MockDataset(ref=ref.to_simple())
                inputs_list.append(input_dataset)
            mock_dataset_quantum.inputs[name] = inputs_list

        # store mock outputs
        for name, refs in outputRefs:
            if not isinstance(refs, list):
                refs = [refs]
            for ref in refs:
                output = MockDataset(
                    ref=ref.to_simple(), quantum=mock_dataset_quantum, output_connection_name=name
                )
                butlerQC.put(output, ref)

        _LOG.info("Finished mocking task '%s' on quantum %s", self.getName(), quantum.dataId)



# This file is part of pipe_base.
#
# Developed for the LSST Data Management System.
# This product includes software developed by the LSST Project
# (http://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
from __future__ import annotations

__all__ = ("_DatasetTracker", "DatasetTypeName", "_pruner")

from collections import defaultdict
from collections.abc import Iterable
from itertools import chain
from typing import Generic, NewType, TypeVar

import networkx as nx
from lsst.daf.butler import DatasetRef, DatasetType, NamedKeyDict, Quantum
from lsst.pipe.base.connections import AdjustQuantumHelper

from .._status import NoWorkFound
from ..pipeline import TaskDef
from .quantumNode import QuantumNode

# NewTypes
DatasetTypeName = NewType("DatasetTypeName", str)

# Generic type parameters
_T = TypeVar("_T", DatasetTypeName, DatasetRef)
_U = TypeVar("_U", TaskDef, QuantumNode)


class _DatasetTracker(Generic[_T, _U]):
    r"""A generic container for tracking keys which are produced or
    consumed by some value. In the context of a QuantumGraph, keys may be
    `~lsst.daf.butler.DatasetRef`\ s and the values would be Quanta that either
    produce or consume those `~lsst.daf.butler.DatasetRef`\ s.

    Prameters
    ---------
    createInverse : bool
        When adding a key associated with a producer or consumer, also create
        and inverse mapping that allows looking up all the keys associated with
        some value. Defaults to False.
    """

    def __init__(self, createInverse: bool = False):
        self._producers: dict[_T, _U] = {}
        self._consumers: defaultdict[_T, set[_U]] = defaultdict(set)
        self._createInverse = createInverse
        if self._createInverse:
            self._itemsDict: defaultdict[_U, set[_T]] = defaultdict(set)

    def addProducer(self, key: _T, value: _U) -> None:
        """Add a key which is produced by some value.

        Parameters
        ----------
        key : `~typing.TypeVar`
            The type to track.
        value : `~typing.TypeVar`
            The type associated with the production of the key.

        Raises
        ------
        ValueError
            Raised if key is already declared to be produced by another value.
        """
        if (existing := self._producers.get(key)) is not None and existing != value:
            raise ValueError(f"Only one node is allowed to produce {key}, the current producer is {existing}")
        self._producers[key] = value
        if self._createInverse:
            self._itemsDict[value].add(key)

    def removeProducer(self, key: _T, value: _U) -> None:
        """Remove a value (e.g. `QuantumNode` or `TaskDef`) from being
        considered a producer of the corresponding key.

        It is not an error to remove a key that is not in the tracker.

        Parameters
        ----------
        key : `~typing.TypeVar`
            The type to track.
        value : `~typing.TypeVar`
            The type associated with the production of the key.
        """
        self._producers.pop(key, None)
        if self._createInverse:
            if result := self._itemsDict.get(value):
                result.discard(key)

    def addConsumer(self, key: _T, value: _U) -> None:
        """Add a key which is consumed by some value.

        Parameters
        ----------
        key : `~typing.TypeVar`
            The type to track.
        value : `~typing.TypeVar`
            The type associated with the consumption of the key.
        """
        self._consumers[key].add(value)
        if self._createInverse:
            self._itemsDict[value].add(key)

    def removeConsumer(self, key: _T, value: _U) -> None:
        """Remove a value (e.g. `QuantumNode` or `TaskDef`) from being
        considered a consumer of the corresponding key.

        It is not an error to remove a key that is not in the tracker.

        Parameters
        ----------
        key : `~typing.TypeVar`
            The type to track.
        value : `~typing.TypeVar`
            The type associated with the consumption of the key.
        """
        if (result := self._consumers.get(key)) is not None:
            result.discard(value)
        if self._createInverse:
            if result_inverse := self._itemsDict.get(value):
                result_inverse.discard(key)

    def getConsumers(self, key: _T) -> set[_U]:
        """Return all values associated with the consumption of the supplied
        key.

        Parameters
        ----------
        key : `~typing.TypeVar`
            The type which has been tracked in the `_DatasetTracker`.
        """
        return self._consumers.get(key, set())

    def getProducer(self, key: _T) -> _U | None:
        """Return the value associated with the consumption of the supplied
        key.

        Parameters
        ----------
        key : `~typing.TypeVar`
            The type which has been tracked in the `_DatasetTracker`.
        """
        # This tracker may have had all nodes associated with a key removed
        # and if there are no refs (empty set) should return None
        return producer if (producer := self._producers.get(key)) else None

    def getAll(self, key: _T) -> set[_U]:
        """Return all consumers and the producer associated with the the
        supplied key.

        Parameters
        ----------
        key : `~typing.TypeVar`
            The type which has been tracked in the `_DatasetTracker`.
        """
        return self.getConsumers(key).union(x for x in (self.getProducer(key),) if x is not None)

    @property
    def inverse(self) -> defaultdict[_U, set[_T]] | None:
        """Return the inverse mapping if class was instantiated to create an
        inverse, else return None.
        """
        return self._itemsDict if self._createInverse else None

    def makeNetworkXGraph(self) -> nx.DiGraph:
        """Create a NetworkX graph out of all the contained keys, using the
        relations of producer and consumers to create the edges.

        Returns
        -------
        graph : `networkx.DiGraph`
            The graph created out of the supplied keys and their relations.
        """
        graph = nx.DiGraph()
        for entry in self._producers.keys() | self._consumers.keys():
            producer = self.getProducer(entry)
            consumers = self.getConsumers(entry)
            # This block is for tasks that consume existing inputs
            if producer is None and consumers:
                for consumer in consumers:
                    graph.add_node(consumer)
            # This block is for tasks that produce output that is not consumed
            # in this graph
            elif producer is not None and not consumers:
                graph.add_node(producer)
            # all other connections
            else:
                for consumer in consumers:
                    graph.add_edge(producer, consumer)
        return graph

    def keys(self) -> set[_T]:
        """Return all tracked keys."""
        return self._producers.keys() | self._consumers.keys()

    def remove(self, key: _T) -> None:
        """Remove a key and its corresponding value from the tracker, this is
        a no-op if the key is not in the tracker.

        Parameters
        ----------
        key : `~typing.TypeVar`
            A key tracked by the `_DatasetTracker`.
        """
        self._producers.pop(key, None)
        self._consumers.pop(key, None)

    def __contains__(self, key: _T) -> bool:
        """Check if a key is in the `_DatasetTracker`.

        Parameters
        ----------
        key : `~typing.TypeVar`
            The key to check.

        Returns
        -------
        contains : `bool`
            Boolean of the presence of the supplied key.
        """
        return key in self._producers or key in self._consumers


def _pruner(
    datasetRefDict: _DatasetTracker[DatasetRef, QuantumNode],
    refsToRemove: Iterable[DatasetRef],
    *,
    alreadyPruned: set[QuantumNode] | None = None,
) -> None:
    r"""Prune supplied dataset refs out of ``datasetRefDict`` container,
    recursing to additional nodes dependant on pruned refs.

    Parameters
    ----------
    datasetRefDict : `_DatasetTracker` [ `~lsst.daf.butler.DatasetRef`, \
            `QuantumNode`]
        The dataset tracker that maps `~lsst.daf.butler.DatasetRef`\ s to the
        `QuantumNode`\s that produce/consume that
        `~lsst.daf.butler.DatasetRef`.
        This function modifies ``datasetRefDict`` in-place.
    refsToRemove : `~collections.abc.Iterable` of `~lsst.daf.butler.DatasetRef`
        The `~lsst.daf.butler.DatasetRef`\ s which should be pruned from the
        input dataset tracker.
    alreadyPruned : `set` of `QuantumNode`
        A set of nodes which have been pruned from the dataset tracker.
    """
    if alreadyPruned is None:
        alreadyPruned = set()
    for ref in refsToRemove:
        # make a copy here, because this structure will be modified in
        # recursion, hitting a node more than once won't be much of an
        # issue, as we skip anything that has been processed
        nodes = set(datasetRefDict.getConsumers(ref))
        for node in nodes:
            # This node will never be associated with this ref
            datasetRefDict.removeConsumer(ref, node)
            if node in alreadyPruned:
                continue
            # find the connection corresponding to the input ref
            connectionRefs = node.quantum.inputs.get(ref.datasetType)
            if connectionRefs is None:
                # look to see if any inputs are component refs that match the
                # input ref to prune
                others = ref.datasetType.makeAllComponentDatasetTypes()
                # for each other component type check if there are assocated
                # refs
                for other in others:
                    connectionRefs = node.quantum.inputs.get(other)
                    if connectionRefs is not None:
                        # now search the component refs and see which one
                        # matches the ref to trim
                        for cr in connectionRefs:
                            if cr.makeCompositeRef() == ref:
                                toRemove = cr
                        break
                else:
                    # Ref must be an initInput ref and we want to ignore those
                    raise RuntimeError(f"Cannot prune on non-Input dataset type {ref.datasetType.name}")
            else:
                toRemove = ref

            tmpRefs = set(connectionRefs).difference((toRemove,))
            tmpConnections = NamedKeyDict[DatasetType, list[DatasetRef]](node.quantum.inputs.items())
            tmpConnections[toRemove.datasetType] = list(tmpRefs)
            helper = AdjustQuantumHelper(inputs=tmpConnections, outputs=node.quantum.outputs)
            assert node.quantum.dataId is not None, (
                "assert to make the type checker happy, it should not "
                "actually be possible to not have dataId set to None "
                "at this point"
            )

            # Try to adjust the quantum with the reduced refs to make sure the
            # node will still satisfy all its conditions.
            #
            # If it can't because NoWorkFound is raised, that means a
            # connection is no longer present, and the node should be removed
            # from the graph.
            try:
                helper.adjust_in_place(node.taskDef.connections, node.taskDef.label, node.quantum.dataId)
                newQuantum = Quantum(
                    taskName=node.quantum.taskName,
                    taskClass=node.quantum.taskClass,
                    dataId=node.quantum.dataId,
                    initInputs=node.quantum.initInputs,
                    inputs=helper.inputs,
                    outputs=helper.outputs,
                )
                # If the inputs or outputs were adjusted to something different
                # than what was supplied by the graph builder, dissassociate
                # node from those refs, and if they are output refs, prune them
                # from downstream tasks. This means that based on new inputs
                # the task wants to produce fewer outputs, or consume fewer
                # inputs.
                for condition, existingMapping, newMapping, remover in (
                    (
                        helper.inputs_adjusted,
                        node.quantum.inputs,
                        helper.inputs,
                        datasetRefDict.removeConsumer,
                    ),
                    (
                        helper.outputs_adjusted,
                        node.quantum.outputs,
                        helper.outputs,
                        datasetRefDict.removeProducer,
                    ),
                ):
                    if condition:
                        notNeeded = set()
                        for key in existingMapping:
                            if key not in newMapping:
                                compositeRefs = (
                                    r if not r.isComponent() else r.makeCompositeRef()
                                    for r in existingMapping[key]
                                )
                                notNeeded |= set(compositeRefs)
                                continue
                            notNeeded |= set(existingMapping[key]) - set(newMapping[key])
                        if notNeeded:
                            for ref in notNeeded:
                                if ref.isComponent():
                                    ref = ref.makeCompositeRef()
                                remover(ref, node)
                            if remover is datasetRefDict.removeProducer:
                                _pruner(datasetRefDict, notNeeded, alreadyPruned=alreadyPruned)
                object.__setattr__(node, "quantum", newQuantum)
                noWorkFound = False

            except NoWorkFound:
                noWorkFound = True

            if noWorkFound:
                # This will throw if the length is less than the minimum number
                for tmpRef in chain(
                    chain.from_iterable(node.quantum.inputs.values()), node.quantum.initInputs.values()
                ):
                    if tmpRef.isComponent():
                        tmpRef = tmpRef.makeCompositeRef()
                    datasetRefDict.removeConsumer(tmpRef, node)
                alreadyPruned.add(node)
                # prune all outputs produced by this node
                # mark that none of these will be produced
                forwardPrunes = set()
                for forwardRef in chain.from_iterable(node.quantum.outputs.values()):
                    datasetRefDict.removeProducer(forwardRef, node)
                    forwardPrunes.add(forwardRef)
                _pruner(datasetRefDict, forwardPrunes, alreadyPruned=alreadyPruned)



# This file is part of pipe_base.
#
# Developed for the LSST Data Management System.
# This product includes software developed by the LSST Project
# (https://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from __future__ import annotations

__all__ = [
    "assertValidInitOutput",
    "assertValidOutput",
    "getInitInputs",
    "lintConnections",
    "makeQuantum",
    "runTestQuantum",
]


import collections.abc
import itertools
import unittest.mock
from collections import defaultdict
from collections.abc import Mapping, Sequence, Set
from typing import TYPE_CHECKING, Any

from lsst.daf.butler import (
    Butler,
    DataCoordinate,
    DataId,
    DatasetRef,
    DatasetType,
    Dimension,
    DimensionUniverse,
    Quantum,
    SkyPixDimension,
    StorageClassFactory,
)
from lsst.pipe.base.connectionTypes import BaseConnection, DimensionedConnection

from .butlerQuantumContext import ButlerQuantumContext

if TYPE_CHECKING:
    from .config import PipelineTaskConfig
    from .connections import PipelineTaskConnections
    from .pipelineTask import PipelineTask
    from .struct import Struct


def makeQuantum(
    task: PipelineTask,
    butler: Butler,
    dataId: DataId,
    ioDataIds: Mapping[str, DataId | Sequence[DataId]],
) -> Quantum:
    """Create a Quantum for a particular data ID(s).

    Parameters
    ----------
    task : `lsst.pipe.base.PipelineTask`
        The task whose processing the quantum represents.
    butler : `lsst.daf.butler.Butler`
        The collection the quantum refers to.
    dataId: any data ID type
        The data ID of the quantum. Must have the same dimensions as
        ``task``'s connections class.
    ioDataIds : `collections.abc.Mapping` [`str`]
        A mapping keyed by input/output names. Values must be data IDs for
        single connections and sequences of data IDs for multiple connections.

    Returns
    -------
    quantum : `lsst.daf.butler.Quantum`
        A quantum for ``task``, when called with ``dataIds``.
    """
    # This is a type ignore, because `connections` is a dynamic class, but
    # it for sure will have this property
    connections = task.config.ConnectionsClass(config=task.config)  # type: ignore

    try:
        _checkDimensionsMatch(butler.dimensions, connections.dimensions, dataId.keys())
    except ValueError as e:
        raise ValueError("Error in quantum dimensions.") from e

    inputs = defaultdict(list)
    outputs = defaultdict(list)
    for name in itertools.chain(connections.inputs, connections.prerequisiteInputs):
        try:
            connection = connections.__getattribute__(name)
            _checkDataIdMultiplicity(name, ioDataIds[name], connection.multiple)
            ids = _normalizeDataIds(ioDataIds[name])
            for id in ids:
                ref = _refFromConnection(butler, connection, id)
                inputs[ref.datasetType].append(ref)
        except (ValueError, KeyError) as e:
            raise ValueError(f"Error in connection {name}.") from e
    for name in connections.outputs:
        try:
            connection = connections.__getattribute__(name)
            _checkDataIdMultiplicity(name, ioDataIds[name], connection.multiple)
            ids = _normalizeDataIds(ioDataIds[name])
            for id in ids:
                ref = _refFromConnection(butler, connection, id)
                outputs[ref.datasetType].append(ref)
        except (ValueError, KeyError) as e:
            raise ValueError(f"Error in connection {name}.") from e
    quantum = Quantum(
        taskClass=type(task),
        dataId=DataCoordinate.standardize(dataId, universe=butler.dimensions),
        inputs=inputs,
        outputs=outputs,
    )
    return quantum


def _checkDimensionsMatch(
    universe: DimensionUniverse,
    expected: Set[str] | Set[Dimension],
    actual: Set[str] | Set[Dimension],
) -> None:
    """Test whether two sets of dimensions agree after conversions.

    Parameters
    ----------
    universe : `lsst.daf.butler.DimensionUniverse`
        The set of all known dimensions.
    expected : `Set` [`str`] or `Set` [`~lsst.daf.butler.Dimension`]
        The dimensions expected from a task specification.
    actual : `Set` [`str`] or `Set` [`~lsst.daf.butler.Dimension`]
        The dimensions provided by input.

    Raises
    ------
    ValueError
        Raised if ``expected`` and ``actual`` cannot be reconciled.
    """
    if _simplify(universe, expected) != _simplify(universe, actual):
        raise ValueError(f"Mismatch in dimensions; expected {expected} but got {actual}.")


def _simplify(universe: DimensionUniverse, dimensions: Set[str] | Set[Dimension]) -> set[str]:
    """Reduce a set of dimensions to a string-only form.

    Parameters
    ----------
    universe : `lsst.daf.butler.DimensionUniverse`
        The set of all known dimensions.
    dimensions : `Set` [`str`] or `Set` [`~lsst.daf.butler.Dimension`]
        A set of dimensions to simplify.

    Returns
    -------
    dimensions : `Set` [`str`]
        A copy of ``dimensions`` reduced to string form, with all spatial
        dimensions simplified to ``skypix``.
    """
    simplified: set[str] = set()
    for dimension in dimensions:
        # skypix not a real Dimension, handle it first
        if dimension == "skypix":
            simplified.add(dimension)  # type: ignore
        else:
            # Need a Dimension to test spatialness
            fullDimension = universe[dimension] if isinstance(dimension, str) else dimension
            if isinstance(fullDimension, SkyPixDimension):
                simplified.add("skypix")
            else:
                simplified.add(fullDimension.name)
    return simplified


def _checkDataIdMultiplicity(name: str, dataIds: DataId | Sequence[DataId], multiple: bool) -> None:
    """Test whether data IDs are scalars for scalar connections and sequences
    for multiple connections.

    Parameters
    ----------
    name : `str`
        The name of the connection being tested.
    dataIds : any data ID type or `~collections.abc.Sequence` [data ID]
        The data ID(s) provided for the connection.
    multiple : `bool`
        The ``multiple`` field of the connection.

    Raises
    ------
    ValueError
        Raised if ``dataIds`` and ``multiple`` do not match.
    """
    if multiple:
        if not isinstance(dataIds, collections.abc.Sequence):
            raise ValueError(f"Expected multiple data IDs for {name}, got {dataIds}.")
    else:
        # DataCoordinate is a Mapping
        if not isinstance(dataIds, collections.abc.Mapping):
            raise ValueError(f"Expected single data ID for {name}, got {dataIds}.")


def _normalizeDataIds(dataIds: DataId | Sequence[DataId]) -> Sequence[DataId]:
    """Represent both single and multiple data IDs as a list.

    Parameters
    ----------
    dataIds : any data ID type or `~collections.abc.Sequence` thereof
        The data ID(s) provided for a particular input or output connection.

    Returns
    -------
    normalizedIds : `~collections.abc.Sequence` [data ID]
        A sequence equal to ``dataIds`` if it was already a sequence, or
        ``[dataIds]`` if it was a single ID.
    """
    if isinstance(dataIds, collections.abc.Sequence):
        return dataIds
    else:
        return [dataIds]


def _refFromConnection(
    butler: Butler, connection: DimensionedConnection, dataId: DataId, **kwargs: Any
) -> DatasetRef:
    """Create a DatasetRef for a connection in a collection.

    Parameters
    ----------
    butler : `lsst.daf.butler.Butler`
        The collection to point to.
    connection : `lsst.pipe.base.connectionTypes.DimensionedConnection`
        The connection defining the dataset type to point to.
    dataId
        The data ID for the dataset to point to.
    **kwargs
        Additional keyword arguments used to augment or construct
        a `~lsst.daf.butler.DataCoordinate`.

    Returns
    -------
    ref : `lsst.daf.butler.DatasetRef`
        A reference to a dataset compatible with ``connection``, with ID
        ``dataId``, in the collection pointed to by ``butler``.
    """
    universe = butler.dimensions
    # DatasetRef only tests if required dimension is missing, but not extras
    _checkDimensionsMatch(universe, set(connection.dimensions), dataId.keys())
    dataId = DataCoordinate.standardize(dataId, **kwargs, universe=universe)

    datasetType = butler.registry.getDatasetType(connection.name)

    try:
        butler.registry.getDatasetType(datasetType.name)
    except KeyError:
        raise ValueError(f"Invalid dataset type {connection.name}.")
    if not butler.run:
        raise ValueError("Can not create a resolved DatasetRef since the butler has no default run defined.")
    try:
        registry_ref = butler.registry.findDataset(datasetType, dataId, collections=[butler.run])
        if registry_ref:
            ref = registry_ref
        else:
            ref = DatasetRef(datasetType=datasetType, dataId=dataId, run=butler.run)
            butler.registry._importDatasets([ref])
        return ref
    except KeyError as e:
        raise ValueError(f"Dataset type ({connection.name}) and ID {dataId.byName()} not compatible.") from e


def runTestQuantum(
    task: PipelineTask, butler: Butler, quantum: Quantum, mockRun: bool = True
) -> unittest.mock.Mock | None:
    """Run a PipelineTask on a Quantum.

    Parameters
    ----------
    task : `lsst.pipe.base.PipelineTask`
        The task to run on the quantum.
    butler : `lsst.daf.butler.Butler`
        The collection to run on.
    quantum : `lsst.daf.butler.Quantum`
        The quantum to run.
    mockRun : `bool`
        Whether or not to replace ``task``'s ``run`` method. The default of
        `True` is recommended unless ``run`` needs to do real work (e.g.,
        because the test needs real output datasets).

    Returns
    -------
    run : `unittest.mock.Mock` or `None`
        If ``mockRun`` is set, the mock that replaced ``run``. This object can
        be queried for the arguments ``runQuantum`` passed to ``run``.
    """
    butlerQc = ButlerQuantumContext(butler, quantum)
    # This is a type ignore, because `connections` is a dynamic class, but
    # it for sure will have this property
    connections = task.config.ConnectionsClass(config=task.config)  # type: ignore
    inputRefs, outputRefs = connections.buildDatasetRefs(quantum)
    if mockRun:
        with unittest.mock.patch.object(task, "run") as mock, unittest.mock.patch(
            "lsst.pipe.base.ButlerQuantumContext.put"
        ):
            task.runQuantum(butlerQc, inputRefs, outputRefs)
            return mock
    else:
        task.runQuantum(butlerQc, inputRefs, outputRefs)
        return None


def _assertAttributeMatchesConnection(obj: Any, attrName: str, connection: BaseConnection) -> None:
    """Test that an attribute on an object matches the specification given in
    a connection.

    Parameters
    ----------
    obj
        An object expected to contain the attribute ``attrName``.
    attrName : `str`
        The name of the attribute to be tested.
    connection : `lsst.pipe.base.connectionTypes.BaseConnection`
        The connection, usually some type of output, specifying ``attrName``.

    Raises
    ------
    AssertionError:
        Raised if ``obj.attrName`` does not match what's expected
        from ``connection``.
    """
    # name
    try:
        attrValue = obj.__getattribute__(attrName)
    except AttributeError:
        raise AssertionError(f"No such attribute on {obj!r}: {attrName}")
    # multiple
    if connection.multiple:
        if not isinstance(attrValue, collections.abc.Sequence):
            raise AssertionError(f"Expected {attrName} to be a sequence, got {attrValue!r} instead.")
    else:
        # use lazy evaluation to not use StorageClassFactory unless
        # necessary
        if isinstance(attrValue, collections.abc.Sequence) and not issubclass(
            StorageClassFactory().getStorageClass(connection.storageClass).pytype, collections.abc.Sequence
        ):
            raise AssertionError(f"Expected {attrName} to be a single value, got {attrValue!r} instead.")
    # no test for storageClass, as I'm not sure how much persistence
    # depends on duck-typing


def assertValidOutput(task: PipelineTask, result: Struct) -> None:
    """Test that the output of a call to ``run`` conforms to its own
    connections.

    Parameters
    ----------
    task : `lsst.pipe.base.PipelineTask`
        The task whose connections need validation. This is a fully-configured
        task object to support features such as optional outputs.
    result : `lsst.pipe.base.Struct`
        A result object produced by calling ``task.run``.

    Raises
    ------
    AssertionError:
        Raised if ``result`` does not match what's expected from ``task's``
        connections.
    """
    # This is a type ignore, because `connections` is a dynamic class, but
    # it for sure will have this property
    connections = task.config.ConnectionsClass(config=task.config)  # type: ignore

    for name in connections.outputs:
        connection = connections.__getattribute__(name)
        _assertAttributeMatchesConnection(result, name, connection)


def assertValidInitOutput(task: PipelineTask) -> None:
    """Test that a constructed task conforms to its own init-connections.

    Parameters
    ----------
    task : `lsst.pipe.base.PipelineTask`
        The task whose connections need validation.

    Raises
    ------
    AssertionError:
        Raised if ``task`` does not have the state expected from ``task's``
        connections.
    """
    # This is a type ignore, because `connections` is a dynamic class, but
    # it for sure will have this property
    connections = task.config.ConnectionsClass(config=task.config)  # type: ignore

    for name in connections.initOutputs:
        connection = connections.__getattribute__(name)
        _assertAttributeMatchesConnection(task, name, connection)


def getInitInputs(butler: Butler, config: PipelineTaskConfig) -> dict[str, Any]:
    """Return the initInputs object that would have been passed to a
    `~lsst.pipe.base.PipelineTask` constructor.

    Parameters
    ----------
    butler : `lsst.daf.butler.Butler`
        The repository to search for input datasets. Must have
        pre-configured collections.
    config : `lsst.pipe.base.PipelineTaskConfig`
        The config for the task to be constructed.

    Returns
    -------
    initInputs : `dict` [`str`]
        A dictionary of objects in the format of the ``initInputs`` parameter
        to `lsst.pipe.base.PipelineTask`.
    """
    connections = config.connections.ConnectionsClass(config=config)
    initInputs = {}
    for name in connections.initInputs:
        attribute = getattr(connections, name)
        # Get full dataset type to check for consistency problems
        dsType = DatasetType(attribute.name, butler.dimensions.extract(set()), attribute.storageClass)
        # All initInputs have empty data IDs
        initInputs[name] = butler.get(dsType)

    return initInputs


def lintConnections(
    connections: PipelineTaskConnections,
    *,
    checkMissingMultiple: bool = True,
    checkUnnecessaryMultiple: bool = True,
) -> None:
    """Inspect a connections class for common errors.

    These tests are designed to detect misuse of connections features in
    standard designs. An unusually designed connections class may trigger
    alerts despite being correctly written; specific checks can be turned off
    using keywords.

    Parameters
    ----------
    connections : `lsst.pipe.base.PipelineTaskConnections`-type
        The connections class to test.
    checkMissingMultiple : `bool`
        Whether to test for single connections that would match multiple
        datasets at run time.
    checkUnnecessaryMultiple : `bool`
        Whether to test for multiple connections that would only match
        one dataset.

    Raises
    ------
    AssertionError
        Raised if any of the selected checks fail for any connection.
    """
    # Since all comparisons are inside the class, don't bother
    # normalizing skypix.
    quantumDimensions = connections.dimensions

    errors = ""
    # connectionTypes.DimensionedConnection is implementation detail,
    # don't use it.
    for name in itertools.chain(connections.inputs, connections.prerequisiteInputs, connections.outputs):
        connection: DimensionedConnection = connections.allConnections[name]  # type: ignore
        connDimensions = set(connection.dimensions)
        if checkMissingMultiple and not connection.multiple and connDimensions > quantumDimensions:
            errors += (
                f"Connection {name} may be called with multiple values of "
                f"{connDimensions - quantumDimensions} but has multiple=False.\n"
            )
        if checkUnnecessaryMultiple and connection.multiple and connDimensions <= quantumDimensions:
            errors += (
                f"Connection {name} has multiple=True but can only be called with one "
                f"value of {connDimensions} for each {quantumDimensions}.\n"
            )
    if errors:
        raise AssertionError(errors)



from . import automatic_connection_constants, connectionTypes, pipelineIR
from ._dataset_handle import *
from ._instrument import *
from ._observation_dimension_packer import *
from ._status import *
from ._task_metadata import *
from .butlerQuantumContext import *
from .config import *
from .connections import *
from .executionButlerBuilder import *
from .graph import *
from .graphBuilder import *
from .pipeline import *
from .pipelineTask import *
from .struct import *
from .task import *
from .taskFactory import *
from .version import *



# This file is part of pipe_base.
#
# Developed for the LSST Data Management System.
# This product includes software developed by the LSST Project
# (https://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

__all__ = ["TaskMetadata"]

import itertools
import numbers
import warnings
from collections.abc import Collection, Iterator, Mapping, Sequence
from typing import Any, Protocol

from pydantic import BaseModel, Field, StrictBool, StrictFloat, StrictInt, StrictStr

_DEPRECATION_REASON = "Will be removed after v25."
_DEPRECATION_VERSION = "v24"

# The types allowed in a Task metadata field are restricted
# to allow predictable serialization.
_ALLOWED_PRIMITIVE_TYPES = (str, float, int, bool)


class PropertySetLike(Protocol):
    """Protocol that looks like a ``lsst.daf.base.PropertySet``

    Enough of the API is specified to support conversion of a
    ``PropertySet`` to a `TaskMetadata`.
    """

    def paramNames(self, topLevelOnly: bool = True) -> Collection[str]:
        ...

    def getArray(self, name: str) -> Any:
        ...


def _isListLike(v: Any) -> bool:
    return isinstance(v, Sequence) and not isinstance(v, str)


class TaskMetadata(BaseModel):
    """Dict-like object for storing task metadata.

    Metadata can be stored at two levels: single task or task plus subtasks.
    The later is called full metadata of a task and has a form

        topLevelTaskName:subtaskName:subsubtaskName.itemName

    Metadata item key of a task (`itemName` above) must not contain `.`,
    which serves as a separator in full metadata keys and turns
    the value into sub-dictionary. Arbitrary hierarchies are supported.
    """

    scalars: dict[str, StrictFloat | StrictInt | StrictBool | StrictStr] = Field(default_factory=dict)
    arrays: dict[str, list[StrictFloat] | list[StrictInt] | list[StrictBool] | list[StrictStr]] = Field(
        default_factory=dict
    )
    metadata: dict[str, "TaskMetadata"] = Field(default_factory=dict)

    @classmethod
    def from_dict(cls, d: Mapping[str, Any]) -> "TaskMetadata":
        """Create a TaskMetadata from a dictionary.

        Parameters
        ----------
        d : `~collections.abc.Mapping`
            Mapping to convert. Can be hierarchical. Any dictionaries
            in the hierarchy are converted to `TaskMetadata`.

        Returns
        -------
        meta : `TaskMetadata`
            Newly-constructed metadata.
        """
        metadata = cls()
        for k, v in d.items():
            metadata[k] = v
        return metadata

    @classmethod
    def from_metadata(cls, ps: PropertySetLike) -> "TaskMetadata":
        """Create a TaskMetadata from a PropertySet-like object.

        Parameters
        ----------
        ps : `PropertySetLike` or `TaskMetadata`
            A ``PropertySet``-like object to be transformed to a
            `TaskMetadata`. A `TaskMetadata` can be copied using this
            class method.

        Returns
        -------
        tm : `TaskMetadata`
            Newly-constructed metadata.

        Notes
        -----
        Items stored in single-element arrays in the supplied object
        will be converted to scalars in the newly-created object.
        """
        # Use hierarchical names to assign values from input to output.
        # This API exists for both PropertySet and TaskMetadata.
        # from_dict() does not work because PropertySet is not declared
        # to be a Mapping.
        # PropertySet.toDict() is not present in TaskMetadata so is best
        # avoided.
        metadata = cls()
        for key in sorted(ps.paramNames(topLevelOnly=False)):
            value = ps.getArray(key)
            if len(value) == 1:
                value = value[0]
            metadata[key] = value
        return metadata

    def to_dict(self) -> dict[str, Any]:
        """Convert the class to a simple dictionary.

        Returns
        -------
        d : `dict`
            Simple dictionary that can contain scalar values, array values
            or other dictionary values.

        Notes
        -----
        Unlike `dict()`, this method hides the model layout and combines
        scalars, arrays, and other metadata in the same dictionary. Can be
        used when a simple dictionary is needed.  Use
        `TaskMetadata.from_dict()` to convert it back.
        """
        d: dict[str, Any] = {}
        d.update(self.scalars)
        d.update(self.arrays)
        for k, v in self.metadata.items():
            d[k] = v.to_dict()
        return d

    def add(self, name: str, value: Any) -> None:
        """Store a new value, adding to a list if one already exists.

        Parameters
        ----------
        name : `str`
            Name of the metadata property.
        value
            Metadata property value.
        """
        keys = self._getKeys(name)
        key0 = keys.pop(0)
        if len(keys) == 0:
            # If add() is being used, always store the value in the arrays
            # property as a list. It's likely there will be another call.
            slot_type, value = self._validate_value(value)
            if slot_type == "array":
                pass
            elif slot_type == "scalar":
                value = [value]
            else:
                raise ValueError("add() can only be used for primitive types or sequences of those types.")

            if key0 in self.metadata:
                raise ValueError(f"Can not add() to key '{name}' since that is a TaskMetadata")

            if key0 in self.scalars:
                # Convert scalar to array.
                # MyPy should be able to figure out that List[Union[T1, T2]] is
                # compatible with  Union[List[T1], List[T2]] if the list has
                # only one element, but it can't.
                self.arrays[key0] = [self.scalars.pop(key0)]  # type: ignore

            if key0 in self.arrays:
                # Check that the type is not changing.
                if (curtype := type(self.arrays[key0][0])) is not (newtype := type(value[0])):
                    raise ValueError(f"Type mismatch in add() -- currently {curtype} but adding {newtype}")
                self.arrays[key0].extend(value)
            else:
                self.arrays[key0] = value

            return

        self.metadata[key0].add(".".join(keys), value)

    def getScalar(self, key: str) -> str | int | float | bool:
        """Retrieve a scalar item even if the item is a list.

        Parameters
        ----------
        key : `str`
            Item to retrieve.

        Returns
        -------
        value : `str`, `int`, `float`, or `bool`
            Either the value associated with the key or, if the key
            corresponds to a list, the last item in the list.

        Raises
        ------
        KeyError
            Raised if the item is not found.
        """
        # Used in pipe_tasks.
        # getScalar() is the default behavior for __getitem__.
        return self[key]

    def getArray(self, key: str) -> list[Any]:
        """Retrieve an item as a list even if it is a scalar.

        Parameters
        ----------
        key : `str`
            Item to retrieve.

        Returns
        -------
        values : `list` of any
            A list containing the value or values associated with this item.

        Raises
        ------
        KeyError
            Raised if the item is not found.
        """
        keys = self._getKeys(key)
        key0 = keys.pop(0)
        if len(keys) == 0:
            if key0 in self.arrays:
                return self.arrays[key0]
            elif key0 in self.scalars:
                return [self.scalars[key0]]
            elif key0 in self.metadata:
                return [self.metadata[key0]]
            raise KeyError(f"'{key}' not found")

        try:
            return self.metadata[key0].getArray(".".join(keys))
        except KeyError:
            # Report the correct key.
            raise KeyError(f"'{key}' not found") from None

    def names(self, topLevelOnly: bool = True) -> set[str]:
        """Return the hierarchical keys from the metadata.

        Parameters
        ----------
        topLevelOnly  : `bool`
            If true, return top-level keys, otherwise full metadata item keys.

        Returns
        -------
        names : `collection.abc.Set`
            A set of top-level keys or full metadata item keys, including
            the top-level keys.

        Notes
        -----
        Should never be called in new code with ``topLevelOnly`` set to `True`
        -- this is equivalent to asking for the keys and is the default
        when iterating through the task metadata. In this case a deprecation
        message will be issued and the ability will raise an exception
        in a future release.

        When ``topLevelOnly`` is `False` all keys, including those from the
        hierarchy and the top-level hierarchy, are returned.
        """
        if topLevelOnly:
            warnings.warn("Use keys() instead. " + _DEPRECATION_REASON, FutureWarning)
            return set(self.keys())
        else:
            names = set()
            for k, v in self.items():
                names.add(k)  # Always include the current level
                if isinstance(v, TaskMetadata):
                    names.update({k + "." + item for item in v.names(topLevelOnly=topLevelOnly)})
            return names

    def paramNames(self, topLevelOnly: bool) -> set[str]:
        """Return hierarchical names.

        Parameters
        ----------
        topLevelOnly : `bool`
            Control whether only top-level items are returned or items
            from the hierarchy.

        Returns
        -------
        paramNames : `set` of `str`
            If ``topLevelOnly`` is `True`, returns any keys that are not
            part of a hierarchy. If `False` also returns fully-qualified
            names from the hierarchy. Keys associated with the top
            of a hierarchy are never returned.
        """
        # Currently used by the verify package.
        paramNames = set()
        for k, v in self.items():
            if isinstance(v, TaskMetadata):
                if not topLevelOnly:
                    paramNames.update({k + "." + item for item in v.paramNames(topLevelOnly=topLevelOnly)})
            else:
                paramNames.add(k)
        return paramNames

    @staticmethod
    def _getKeys(key: str) -> list[str]:
        """Return the key hierarchy.

        Parameters
        ----------
        key : `str`
            The key to analyze. Can be dot-separated.

        Returns
        -------
        keys : `list` of `str`
            The key hierarchy that has been split on ``.``.

        Raises
        ------
        KeyError
            Raised if the key is not a string.
        """
        try:
            keys = key.split(".")
        except Exception:
            raise KeyError(f"Invalid key '{key}': only string keys are allowed") from None
        return keys

    def keys(self) -> tuple[str, ...]:
        """Return the top-level keys."""
        return tuple(k for k in self)

    def items(self) -> Iterator[tuple[str, Any]]:
        """Yield the top-level keys and values."""
        for k, v in itertools.chain(self.scalars.items(), self.arrays.items(), self.metadata.items()):
            yield (k, v)

    def __len__(self) -> int:
        """Return the number of items."""
        return len(self.scalars) + len(self.arrays) + len(self.metadata)

    # This is actually a Liskov substitution violation, because
    # pydantic.BaseModel says __iter__ should return something else.  But the
    # pydantic docs say to do exactly this to in order to make a mapping-like
    # BaseModel, so that's what we do.
    def __iter__(self) -> Iterator[str]:  # type: ignore
        """Return an iterator over each key."""
        # The order of keys is not preserved since items can move
        # from scalar to array.
        return itertools.chain(iter(self.scalars), iter(self.arrays), iter(self.metadata))

    def __getitem__(self, key: str) -> Any:
        """Retrieve the item associated with the key.

        Parameters
        ----------
        key : `str`
            The key to retrieve. Can be dot-separated hierarchical.

        Returns
        -------
        value : `TaskMetadata`, `float`, `int`, `bool`, `str`
            A scalar value. For compatibility with ``PropertySet``, if the key
             refers to an array, the final element is returned and not the
             array itself.

        Raises
        ------
        KeyError
            Raised if the item is not found.
        """
        keys = self._getKeys(key)
        key0 = keys.pop(0)
        if len(keys) == 0:
            if key0 in self.scalars:
                return self.scalars[key0]
            if key0 in self.metadata:
                return self.metadata[key0]
            if key0 in self.arrays:
                return self.arrays[key0][-1]
            raise KeyError(f"'{key}' not found")
        # Hierarchical lookup so the top key can only be in the metadata
        # property. Trap KeyError and reraise so that the correct key
        # in the hierarchy is reported.
        try:
            # And forward request to that metadata.
            return self.metadata[key0][".".join(keys)]
        except KeyError:
            raise KeyError(f"'{key}' not found") from None

    def get(self, key: str, default: Any = None) -> Any:
        """Retrieve the item associated with the key or a default.

        Parameters
        ----------
        key : `str`
            The key to retrieve. Can be dot-separated hierarchical.
        default
            The value to return if the key doesnot exist.

        Returns
        -------
        value : `TaskMetadata`, `float`, `int`, `bool`, `str`
            A scalar value.  If the key refers to an array, the final element
            is returned and not the array itself; this is consistent with
            `__getitem__` and `PropertySet.get`, but not ``to_dict().get``.
        """
        try:
            return self[key]
        except KeyError:
            return default

    def __setitem__(self, key: str, item: Any) -> None:
        """Store the given item."""
        keys = self._getKeys(key)
        key0 = keys.pop(0)
        if len(keys) == 0:
            slots: dict[str, dict[str, Any]] = {
                "array": self.arrays,
                "scalar": self.scalars,
                "metadata": self.metadata,
            }
            primary: dict[str, Any] | None = None
            slot_type, item = self._validate_value(item)
            primary = slots.pop(slot_type, None)
            if primary is None:
                raise AssertionError(f"Unknown slot type returned from validator: {slot_type}")

            # Assign the value to the right place.
            primary[key0] = item
            for property in slots.values():
                # Remove any other entries.
                property.pop(key0, None)
            return

        # This must be hierarchical so forward to the child TaskMetadata.
        if key0 not in self.metadata:
            self.metadata[key0] = TaskMetadata()
        self.metadata[key0][".".join(keys)] = item

        # Ensure we have cleared out anything with the same name elsewhere.
        self.scalars.pop(key0, None)
        self.arrays.pop(key0, None)

    def __contains__(self, key: str) -> bool:
        """Determine if the key exists."""
        keys = self._getKeys(key)
        key0 = keys.pop(0)
        if len(keys) == 0:
            return key0 in self.scalars or key0 in self.arrays or key0 in self.metadata

        if key0 in self.metadata:
            return ".".join(keys) in self.metadata[key0]
        return False

    def __delitem__(self, key: str) -> None:
        """Remove the specified item.

        Raises
        ------
        KeyError
            Raised if the item is not present.
        """
        keys = self._getKeys(key)
        key0 = keys.pop(0)
        if len(keys) == 0:
            # MyPy can't figure out that this way to combine the types in the
            # tuple is the one that matters, and annotating a local variable
            # helps it out.
            properties: tuple[dict[str, Any], ...] = (self.scalars, self.arrays, self.metadata)
            for property in properties:
                if key0 in property:
                    del property[key0]
                    return
            raise KeyError(f"'{key}' not found'")

        try:
            del self.metadata[key0][".".join(keys)]
        except KeyError:
            # Report the correct key.
            raise KeyError(f"'{key}' not found'") from None

    def _validate_value(self, value: Any) -> tuple[str, Any]:
        """Validate the given value.

        Parameters
        ----------
        value : Any
            Value to check.

        Returns
        -------
        slot_type : `str`
            The type of value given. Options are "scalar", "array", "metadata".
        item : Any
            The item that was given but possibly modified to conform to
            the slot type.

        Raises
        ------
        ValueError
            Raised if the value is not a recognized type.
        """
        # Test the simplest option first.
        value_type = type(value)
        if value_type in _ALLOWED_PRIMITIVE_TYPES:
            return "scalar", value

        if isinstance(value, TaskMetadata):
            return "metadata", value
        if isinstance(value, Mapping):
            return "metadata", self.from_dict(value)

        if _isListLike(value):
            # For model consistency, need to check that every item in the
            # list has the same type.
            value = list(value)

            type0 = type(value[0])
            for i in value:
                if type(i) != type0:
                    raise ValueError(
                        "Type mismatch in supplied list. TaskMetadata requires all"
                        f" elements have same type but see {type(i)} and {type0}."
                    )

            if type0 not in _ALLOWED_PRIMITIVE_TYPES:
                # Must check to see if we got numpy floats or something.
                type_cast: type
                if isinstance(value[0], numbers.Integral):
                    type_cast = int
                elif isinstance(value[0], numbers.Real):
                    type_cast = float
                else:
                    raise ValueError(
                        f"Supplied list has element of type '{type0}'. "
                        "TaskMetadata can only accept primitive types in lists."
                    )

                value = [type_cast(v) for v in value]

            return "array", value

        # Sometimes a numpy number is given.
        if isinstance(value, numbers.Integral):
            value = int(value)
            return "scalar", value
        if isinstance(value, numbers.Real):
            value = float(value)
            return "scalar", value

        raise ValueError(f"TaskMetadata does not support values of type {value!r}.")


# Needed because a TaskMetadata can contain a TaskMetadata.
TaskMetadata.update_forward_refs()



#
# LSST Data Management System
# Copyright 2008-2015 AURA/LSST.
#
# This product includes software developed by the
# LSST Project (http://www.lsst.org/).
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the LSST License Statement and
# the GNU General Public License along with this program.  If not,
# see <https://www.lsstcorp.org/LegalNotices/>.
#
from __future__ import annotations

__all__: list[str] = []

import contextlib

from deprecated.sphinx import deprecated


@deprecated(
    reason="Replaced by lsst.utils.timer.profile().  Will be removed after v26.0",
    version="v25.0",
    category=FutureWarning,
)
@contextlib.contextmanager
def profile(filename, log=None):
    """Context manager for profiling with cProfile.


    Parameters
    ----------
    filename : `str`
        Filename to which to write profile (profiling disabled if `None` or
        empty).
    log : `logging.Logger`, optional
        Log object for logging the profile operations.

    If profiling is enabled, the context manager returns the cProfile.Profile
    object (otherwise it returns None), which allows additional control over
    profiling.  You can obtain this using the "as" clause, e.g.:

    .. code-block:: python

        with profile(filename) as prof:
            runYourCodeHere()

    The output cumulative profile can be printed with a command-line like:

    .. code-block:: bash

        python -c 'import pstats; \
            pstats.Stats("<filename>").sort_stats("cumtime").print_stats(30)'
    """
    if not filename:
        # Nothing to do
        yield
        return
    from cProfile import Profile

    profile = Profile()
    if log is not None:
        log.info("Enabling cProfile profiling")
    profile.enable()
    yield profile
    profile.disable()
    profile.dump_stats(filename)
    if log is not None:
        log.info("cProfile stats written to %s", filename)
